//Doom Graphics Format Compressor Version 0.2 Alpha

#include <stdio.h>
#include <string.h>

struct DGFImage
{
	unsigned short width;
	unsigned short height;
	short xoffset;
	short yoffset;
	unsigned int* pointers;
	unsigned int* newpointers;
	unsigned short** data;
	int* dataLength;
	int* map;

	DGFImage(int width, int height, int xoffset, int yoffset)
	{
		this->width = width;
		this->height = height;
		this->xoffset = xoffset;
		this->yoffset = yoffset;
		this->pointers = new unsigned int[width];
		this->newpointers = new unsigned int[width]; //This will be used during the writing process.
		this->data = new unsigned short*[width];
		this->dataLength = new int[width];
		this->map = new int[width];
		for(int i = 0;i < width;i++)
		{
			this->map[i] = -1;
		}
	}

	int headerLength(int cropped=0)
	{
		return 8 + (width-cropped)*4;
	}
};

bool verbose = false;
bool autocrop = true;
int compressImage(char* file);
bool compareColumn(DGFImage *img, int x1, int x2);

//io functions
unsigned short getByte(FILE *in);
int getShort(FILE *in, bool signd=false);
unsigned int getInteger(FILE *in);
void writeInteger(FILE *out, int value, int length);
void writePost(unsigned short *data, int length, FILE *out, bool skipfirst);

int main(int argc, char* args[])
{
	printf("Doom Graphics Format Compressor                                      v0.2 Alpha\n");
	printf("Created by Blzut3                                            http://BitOwl.com/\n");
	printf("-------------------------------------------------------------------------------\n");
	if(argc == 1)
	{
		printf("Usage: DGFCompress [files...] [parameters]\n");
		printf("-verbose : Prints extra information about the files being compressed.\n");
		printf("-noautocrop : Disables the autocrop feature.\n");
		return 0;
	}
	for(int i = 1;i < argc;i++)
	{
		if(stricmp(args[i], "-verbose") == 0)
			verbose = true;
		else if(stricmp(args[i], "-noautocrop") == 0)
			autocrop = false;
		else
			continue;
		args[i] = NULL;
	}
	int totalSaved = 0;
	for(int i = 1;i < argc;i++) //Go through all listed files and compress them.
	{
		if(args[i] == NULL)
			continue;
		printf("Compressing %s...\n", args[i]);
		totalSaved += compressImage(args[i]);
	}
	printf("Total Space Saved: %d bytes\n", totalSaved);
	return 0;
}

int compressImage(char* file)
{
	FILE *in = fopen(file, "rb");
	if(in == NULL) //Can't find it?
	{
		printf("Error: Unable to open file.\n");
		return 0;
	}
	DGFImage *img = new DGFImage(getShort(in), getShort(in), getShort(in, true), getShort(in, true));
	//read in the pointers
	for(int x = 0;x < img->width;x++)
	{
		img->pointers[x] = getInteger(in);
	}
	if(verbose) //Print "Useless" information
	{
		if(autocrop) printf("Autocropping is on!\n");
		printf("Width: %d	Height: %d	XOffset: %d	YOffset: %d\n", img->width, img->height, img->xoffset, img->yoffset);
		printf("Pointers (#:pos:len):");
	}
	//column data
	int minoffset = 255; //basically highest pixel
	int lowesty = 0; //lowest pixel
	for(int x = 0;x < img->width;x++)
	{
		if(fseek(in, img->pointers[x], SEEK_SET) != 0)
		{
			printf("Error: Could not seek to pointer.\n");
			return 0;
		}
		//collect length
		int length = 0;
		int offset = 0;
		int leftoffaty = 0;
		bool firstloop = true; //something we need to do the first time the loop is executed
		while((offset = getByte(in)) != 0xFF) //collect the column's data
		{
			if(firstloop)
			{
				leftoffaty = offset;
				firstloop = false;
			}
			if(offset < minoffset)
				minoffset = offset;
			int postLength = getByte(in);
			length += postLength + 4;
			leftoffaty = offset+postLength;
			if(fseek(in, postLength+2, SEEK_CUR) != 0)
			{
				printf("Error: Seek Error.\n");
				return 0;
			}
			if(leftoffaty > lowesty)
				lowesty = leftoffaty;
		}
		if(verbose) printf(" %d:%d:%d", x, img->pointers[x], length);
		img->data[x] = new unsigned short[length];
		img->dataLength[x] = length;
		if(length != 0)
		{
			fseek(in, img->pointers[x], SEEK_SET); //go back to the top
			for(int i = 0;i < length;i++) //Returns more reliably then fread was I think
				img->data[x][i] = getByte(in);
		}
		else
		{
			delete img->data[x];
		}
	}
	int trueimageheight = lowesty - minoffset;
	if(verbose) printf("\nVerticle Cropable Space (Uppper:Lower): %d:%d\nMapping (-1 means no remap):", minoffset, img->height-minoffset-trueimageheight);
	if(autocrop) //apply verical cropping changes
	{
		img->yoffset -= minoffset;
		img->height -= img->height-minoffset-trueimageheight;
		img->height -= minoffset;
		for(int x = 0;x < img->width;x++)
		{
			int pos = 0;
			if(img->dataLength[x] != 0) //do we have posts to shift
			{
				do
				{
					img->data[x][pos] -= minoffset;
					pos += img->data[x][pos+1]+4;
				}
				while(img->dataLength[x] > pos);
			}
		}
	}
	bool dataReached = false;
	int croppedwidth = 0;
	for(int x1 = 0;x1 < img->width;x1++)
	{
		if(autocrop) //x axis cropping
		{
			if(!dataReached && img->dataLength[x1] == 0)
			{
				img->map[x1] = -2;
				img->xoffset--;
				croppedwidth++;
				if(verbose) printf(" %d=cropped", x1);
				continue;
			}
			else if(img->dataLength[x1] != 0)
				dataReached = true;
			else
			{
				bool crop = true;
				for(int x2 = x1+1;x2 < img->width;x2++) //see if there is any more data
				{
					if(img->dataLength[x2] != 0)
					{
						crop = false;
						break;
					}
				}
				if(crop)
				{
					for(int x2 = x1;x2 < img->width;x2++) //go through again and mark them all from this point as cropped.
					{
						img->map[x2] = -2;
						croppedwidth++;
						if(verbose) printf(" %d=cropped", x2);
					}
					break; //we are done
				}
			}
		}
		for(int x2 = x1+1;x2 < img->width;x2++)
		{
			if(compareColumn(img, x1, x2))
			{
				if(img->map[x1] == -1)
					img->map[x2] = x1;
				else
					img->map[x2] = img->map[x1];
				break;
			}
		}
		if(verbose) printf(" %d=>%d", x1, img->map[x1]);
	}
	if(verbose) printf("\n");
	//Unlike verticle space this isn't calculated when -noautocrop is provided.
	if(autocrop && verbose) printf("Horizontal Cropable Space (Total): %d\n", croppedwidth);
	//Seek to the end of the file to try to get the file size
	if(fseek(in, 0, SEEK_END) != 0)
	{
		printf("Error: Seek Error.\n");
		return 0;
	}
	long int originalSize = ftell(in);
	fclose(in);
	char* newfilename = new char[sizeof(file)+4];
	sprintf(newfilename, "%s%s", file, ".bak");
	remove(newfilename);
	rename(file, newfilename);
	//now lets rewrite the image
	FILE *out = fopen(file, "wb");
	writeInteger(out, img->width - croppedwidth, 2);
	writeInteger(out, img->height, 2);
	writeInteger(out, img->xoffset, 2);
	writeInteger(out, img->yoffset, 2);
	unsigned int ptr = img->headerLength(croppedwidth);
	for(int x = 0;x < img->width;x++)
	{
		if(img->map[x] == -2) //gone
			continue;
		else if(img->map[x] == -1)
		{
			//if the post offset is (usually) 0 and the length is 255 then save two bytes.
			if(x != 0 && img->data[x][0] == img->data[x-1][img->dataLength[x-1]-1] && img->data[x][1] == 0xFF)
			{ //EXTREMELY RARE
				ptr -= 2;
				if(verbose) printf("Byte saver activated for %d.\n", x);
			}
			writeInteger(out, ptr, 4);
			img->newpointers[x] = ptr;
			ptr++;
			ptr += img->dataLength[x];
		}
		else
		{
			writeInteger(out, img->newpointers[img->map[x]], 4);
			img->newpointers[x] = img->newpointers[img->map[x]];
		}
	}
	for(int x = 0;x < img->width;x++) //go through again to write data
	{
		if(img->map[x] == -1)
		{
			if(img->dataLength[x] != 0)
				writePost(img->data[x], img->dataLength[x], out, (x != 0 && img->data[x][0] == 0xFF) ? true : false);
			writeInteger(out, 0xFF, 1);
		}
	}
	long int newSize = ftell(out);
	fclose(out);
	int spaceSaved = originalSize - newSize;
	if(spaceSaved == 0)
		printf("Unable to compress image.\n");
	else
		printf("Saved %d bytes!\n", spaceSaved);
	return spaceSaved;
}

unsigned short getByte(FILE *in) //Someone please tell me if there is a better way to get single bytes from a file reliably.
{
	return (unsigned short) fgetc(in);
}

int getShort(FILE *in, bool signd)
{
	if(signd)
	{
		short value[1];
		fread(value, 2, 1, in);
		return value[0];
	}
	else
	{
		int value[1];
		fread(value, 2, 1, in);
		return value[0];
	}
}

unsigned int getInteger(FILE *in)
{
	unsigned int value[1];
	fread(value, 4, 1, in);
	return value[0];
}

bool compareColumn(DGFImage *img, int x1, int x2)
{
	unsigned short* c1 = img->data[x1];
	unsigned short* c2 = img->data[x2];
	if(c1 == NULL && c2 == NULL)
		return true;
	else if(c1 == NULL || c2 == NULL)
		return false;
	if(img->dataLength[x1] != img->dataLength[x2])
		return false;
	for(int i = 0;i < img->dataLength[x1];i++)
	{
		if(c1[i] != c2[i])
		{
			return false;
		}
	}
	return true;
}

void writeInteger(FILE *out, int value, int length)
{
	int* val = new int[1];
	val[0] = value;
	fwrite(val, length, 1, out);
}

void writePost(unsigned short *data, int length, FILE *out, bool skipfirst)
{
	if(skipfirst)
	{
		unsigned short *writedata = new unsigned short[length-2];
		for(int i = 2;i < length;i++)
			writedata[i-2] = data[i];
		for(int i = 0;i < length-2;i++)
			writeInteger(out, writedata[i], 1);
	}
	else
	{
		for(int i = 0;i < length;i++)
			writeInteger(out, data[i], 1);
	}
}
